
/*

   muscan - scanner driver for Mustek Paragon 6000CX scanner
   Copyright (C) 1995 Torsten Eichner

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

   email:       eichner@rhrk.uni-kl.de
   papermail:   Torsten Eichner, Kampshardtweg 1, D-57587 Birken-Honigsessen, Germany

 */

/*
   Start of file: muscan.c
 */

#include <errno.h>
#include <fcntl.h>
#include <malloc.h>
#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <unistd.h>

#include </usr/src/linux/drivers/scsi/sg.h>
static char *scanner_device = DEVICE;
static int scanner_fd = -1;

static char *outputfile = "-";
static float x_size = 8.3;
static float y_size = 11.7;
static float x_pos = 0;
static float y_pos = 0;
static int resolution = 300;
static int color = 0;
static int debug = 0;
static int brightness = 12;
static int contrast = 12;
static int grain = 0;
static int speed = 1;
static int backtracking = 1;
static int mem_size = 4096;

static void usage (char *);
static void parse_commandline (int, char *[]);
static void open_scanner (void);
static void close_scanner (void);
static void scan_image (void);
static int scsi_cmd (unsigned char *, int, unsigned char *, int);

#define ERROR(string...) { fprintf(stderr, "ERROR[%s:%d]: ", __FILE__, __LINE__); fprintf(stderr, ##string); exit(-1); }
#define TRACE(lvl, string...) { if (debug >= lvl) { fprintf(stderr, "TRACE[%s:%d]: ", __FILE__, __LINE__); fprintf(stderr, ##string); } }

/*******************************************
 * Print help                              *
 *******************************************/
static void usage
  (
    char *text			/* I: NULL or erroneous argument */
)
{
  printf ("muscan V%s\n", VERSION);

  printf ("Syntax: muscan -b -c -d -f outputfile -h -l pos -r resolution -u pos -v -x length -y length -?\n");
  printf ("Options:\n");
  printf (" -b      : turn off backtracking\n");
  printf (" -c      : use colormode (default is greyscale)\n");
  printf (" -d      : turn on debugging\n");
  printf (" -f file : copy to file instead of stdout\n");
  printf (" -h      : print this help\n");
  printf (" -l pos  : set x position of image (default 0\")\n");
  printf (" -p block: set various parameters (block=brightness:contrast:grain:speed, default 12:12:0:1)\n");
  printf (" -r res  : set resolution to res dpi (default 300)\n");
  printf (" -u pos  : set y position of image (default 0\")\n");
  printf (" -v      : show version\n");
  printf (" -x len  : set x size in inch (default 8.3 = DIN A4\")\n");
  printf (" -y len  : set y size in inch (default 11.7 = DIN A4\")\n");
  printf (" -?      : print this help\n");

  if (text)
    printf ("Erroneous argument: <%s>\n", text);

  if (text)
    exit (-1);
  else
    exit (0);
}

/*******************************************
 * Parse commandline                       *
 *******************************************/
static void parse_commandline
  (
    int argc,			/* I: argument count */
    char *argv[]		/* I: arguments */
)
{
  int i;

  for (i = 1; i < argc; i++)
    if (argv[i][0] == '-')
      switch (argv[i][1])
	{
	case 'b':		/* backtracking */
	  backtracking = 0;
	  break;

	case 'c':		/* color mode */
	  color = 1;
	  break;

	case 'd':		/* debug mode */
	  debug++;
	  break;

	case 'f':		/* outputfile */
	  i++;
	  if (i < argc)
	    outputfile = argv[i];
	  else
	    usage (argv[i - 1]);
	  break;

	case 'h':		/* help */
	  usage (NULL);
	  break;

	case 'l':		/* x-pos */
	  i++;
	  if (i >= argc || sscanf (argv[i], "%f", &x_pos) != 1 || x_pos < 0.0)
	    usage (argv[i - 1]);
	  break;

	case 'p':		/* various parameters */
	  i++;
	  if (i >= argc || sscanf (argv[i], "%d:%d:%d:%d", &brightness, &contrast, &grain, &speed) != 4 || brightness < 0 || contrast < 0 || grain < 0 || speed < 0 || brightness > 24 || contrast > 24 || grain > 5 || speed > 4)
	    usage (argv[i - 1]);
	  break;

	case 'r':		/* resolution */
	  i++;
	  if (i >= argc || sscanf (argv[i], "%d", &resolution) != 1 || resolution < 1 || resolution > 600)
	    usage (argv[i - 1]);
	  break;

	case 'u':		/* y-pos */
	  i++;
	  if (i >= argc || sscanf (argv[i], "%f", &y_pos) != 1 || y_pos < 0.0)
	    usage (argv[i - 1]);
	  break;

	case 'v':		/* version */
	  printf ("muscan V%s - compiled by %s on %s\n", VERSION, USER, HOST);
	  printf ("Copyright (C) 1995 by Torsten Eichner, Kampshardtweg 1, D-57587 Birken-Honigsessen\n");
	  printf ("E-mail: eichner@rhrk.uni-kl.de\n");
	  printf ("This program is distributed under the GNU GPL - please refer to it for further information.\n");
	  break;

	case 'x':		/* x-size */
	  i++;
	  if (i >= argc || sscanf (argv[i], "%f", &x_size) != 1 || x_size < 0.0)
	    usage (argv[i - 1]);
	  break;

	case 'y':		/* y-size */
	  i++;
	  if (i >= argc || sscanf (argv[i], "%f", &y_size) != 1 || y_size < 0.0)
	    usage (argv[i - 1]);
	  break;

	case '?':		/* help */
	  usage (NULL);
	  break;

	default:
	  usage (argv[i]);
	  break;
	}
    else
      usage (argv[i]);

  TRACE (1, "scanning %2.2f\"x%2.2f\" %s-image to <%s> at pos %2.2f\":%2.2f\" with %d dpi\n", x_size, y_size, color ? "color" : "greyscale", outputfile, x_pos, y_pos, resolution);
  TRACE (1, "pblock %d:%d:%d:%d\n", brightness, contrast, grain, speed);
}

/*******************************************
 * Issue a SCSI command                    *
 *******************************************/
static int scsi_cmd
  (
    unsigned char *src,		/* I: source data */
    int src_size,		/* I: length of source data */
    unsigned char *dest,	/* O: result */
    int dest_size		/* I: expected size of result */
)
{
#ifndef SG_BIG_BUFF
  unsigned char scsi_buffer[4096 + sizeof (struct sg_header)];
#else
  unsigned char scsi_buffer[SG_BIG_BUFF];
#endif
  struct sg_header hdr;
  int i, count;

  TRACE (2, "issue SCSI command 0x%02x (expected replylen: %d+%d=%d)\n", src[0], dest_size, sizeof (struct sg_header), dest_size + sizeof (struct sg_header));

#ifdef SG_BIG_BUF
  if (dest_size > SG_BIG_BUF - sizeof (struct sg_header))
      ERROR ("buffer overflow: expected %d bytes, buffer is %d bytes\n", count, SG_BIG_BUF - sizeof (struct sg_header));
#endif

  hdr.pack_len = src_size + sizeof (struct sg_header);
  hdr.reply_len = dest_size + sizeof (struct sg_header);

  for (i = 0; i < sizeof (struct sg_header); i++)
    scsi_buffer[i] = ((char *) &hdr)[i];

  for (; i < hdr.pack_len; i++)
    scsi_buffer[i] = src[i - sizeof (struct sg_header)];

  if (write (scanner_fd, scsi_buffer, hdr.pack_len) != hdr.pack_len)
    ERROR ("error on write: %s\n", strerror (errno));

  if ((count = read (scanner_fd, scsi_buffer, dest_size + sizeof (struct sg_header))) < 0)
      ERROR ("error on read: %s\n", strerror (errno));

  for (i = 0; i < sizeof (struct sg_header); i++)
    ((char *) &hdr)[i] = scsi_buffer[i];

  if (hdr.result != 0)
    ERROR ("SCSI error (%02x): (cmd %02x) sense 0x%02x 0x%02x 0x%02x...\n", hdr.result, src[0], hdr.sense_buffer[0], hdr.sense_buffer[1], hdr.sense_buffer[2]);

  TRACE (2, "SCSI: read %4d packlen %4d replylen %d result 0x%02x\n", count, hdr.pack_len, hdr.reply_len, hdr.result);
  TRACE (2, "SCSI: sense %02x %02x %02x %02x %02x %02x %02x %02x %02x %02x %02x %02x %02x %02x %02x %02x\n", hdr.sense_buffer[0], hdr.sense_buffer[1], hdr.sense_buffer[2], hdr.sense_buffer[3], hdr.sense_buffer[4], hdr.sense_buffer[5], hdr.sense_buffer[6], hdr.sense_buffer[7], hdr.sense_buffer[8], hdr.sense_buffer[9], hdr.sense_buffer[10], hdr.sense_buffer[11], hdr.sense_buffer[12], hdr.sense_buffer[13], hdr.sense_buffer[14], hdr.sense_buffer[15]);

  TRACE (3, "SCSI data (BEGIN)\n");

  for (; i < count; i++)
    {
      dest[i - sizeof (struct sg_header)] = scsi_buffer[i];
      if ((src[0] != 0x08 && debug > 2) || (debug > 3))
	fprintf (stderr, "%4x:%02x(%c)%s", i - sizeof (struct sg_header), scsi_buffer[i], scsi_buffer[i] >= 32 && scsi_buffer[i] < 127 ? scsi_buffer[i] : ' ', (i - sizeof (struct sg_header) + 1) % 8 == 0 ? "\n" : " ");
    }

  TRACE (3, "SCSI data (END)\n");

  return count - sizeof (struct sg_header);
}

/*******************************************
 * Open scanner                            *
 *******************************************/
static void open_scanner
  (
    void
)
{
  unsigned char inquiry[] =
  {0x12, 0x00, 0x00, 0x00, 0xff, 0x00};
  unsigned char result[256];
#ifdef SG_BIG_BUFF
  union
    {
      unsigned char tmp[SG_BIG_BUFF + sizeof (struct sg_header)];
      struct sg_header hdr;
    }
  sgb;
  int i;
#endif
  int valid_size;

  TRACE (1, "opening scanner <%s>\n", scanner_device);

  if ((scanner_fd = open (scanner_device, O_RDWR | O_EXCL)) < 0)
    ERROR ("cannot open scanner <%s> due to: %s\n", scanner_device, strerror (errno));

  scsi_cmd (inquiry, 6, result, 256);

  if (result[0] != 6 || strncmp (result + 8, "MUSTEK", 6))
    ERROR ("no MUSTEK scanner found\n");

#ifdef SG_BIG_BUFF
  /* try to determine maximum buffer size */
  for (valid_size = 4096; valid_size < SG_BIG_BUFF; valid_size += 512)
    {
      sgb.hdr.pack_len = 6 + sizeof (struct sg_header);
      sgb.hdr.reply_len = valid_size;
      for (i = 0; i < 6; i++)
	sgb.tmp[i + sizeof (struct sg_header)] = inquiry[i];
      if (write (scanner_fd, &sgb, 6 + sizeof (struct sg_header)) < 0)
	  break;

      read (scanner_fd, &sgb, sgb.hdr.reply_len + sizeof (struct sg_header));

      mem_size = valid_size;
    }

  TRACE (1, "new mem_size is %d\n", mem_size);
#endif
}

/*******************************************
 * Close scanner                           *
 *******************************************/
static void close_scanner
  (
    void
)
{
  TRACE (1, "closing scanner <%s>\n", scanner_device);

  close (scanner_fd);
}

/*******************************************
 * Scan image                              *
 *******************************************/
static void scan_image
  (
    void
)
{
  unsigned char inquiry[] =
  {0x12, 0x00, 0x00, 0x00, 0xff, 0x00};
  unsigned char testunit[] =
  {0x00, 0x00, 0x00, 0x00, 0x00, 0x00};
  unsigned char area[] =
  {0x04, 0x00, 0x00, 0x00, 0x09, 0x00, 0x01, 0, 0, 0, 0, 0, 0, 0, 0};
  unsigned char mode[] =
  {0x15, 0x00, 0x00, 0x00, 0x0b, 0x00, 0x83, 0x00, 0x0c, 0x0c, 0x00, 0x04, 0x00, 0x00, 0x00, 0x00};
  unsigned char lookup[] =
  {0x55, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00};
  unsigned char start[] =
  {0x1b, 0x00, 0x00, 0x00, 0x00, 0x00};
  unsigned char stop[] =
  {0x1b, 0x00, 0x00, 0x00, 0x00, 0x00};
  unsigned char status[] =
  {0x0f, 0x00, 0x00, 0x00, 0x06, 0x00};
  unsigned char backtrack[] =
  {0x10, 0x00, 0x00, 0x00, 0x00, 0x00};
  unsigned char readline[] =
  {0x08, 0x00, 0x00, 0x00, 0x01, 0x00};
#ifndef SG_BIG_BUFF
  unsigned char image[4096];
#else
  unsigned char image[SG_BIG_BUFF];
#endif
  unsigned char state[6];
  int run, i, x, y, line, maxline, bpl, maxbpl;
  unsigned char *slice = NULL;
  unsigned char *rgb[3] =
  {NULL, NULL, NULL};
  FILE *stream;
  int ll;

#define MAXRETRY 8
#define BIGRETRY 3

  TRACE (1, "scan image\n");

  maxbpl = resolution * x_size;
  maxline = resolution * y_size;

  if (resolution > 300 || resolution < 3)
    ERROR ("need resolution in the range of 3..300\n");

  if (y_size + y_pos > 14.0)
    ERROR ("illegal size/length: would run over border by %2.2f\"\n", y_size + y_pos - 14.0);

  if (x_size + x_pos > 9.0)
    ERROR ("illegal size/width: would run over border by %2.2f\"\n", x_size + x_pos - 9.0);

  if (maxbpl >= 4095)
    ERROR ("line overflow (%d, allowed 4095) - use lower resolution or x-size\n", maxbpl);

  for (run = 0; run < (color ? 3 : 1); run++)
    {
      int busy, lib, retry, bigretry;
      unsigned char mask;

      bigretry = 0;
      retry = 0;

      if ((slice = malloc (maxbpl * maxline)) == NULL)
	ERROR ("cannot malloc %d bytes\n", maxbpl * maxline);

      for (x = 0; x < maxbpl * maxline; x++)
	slice[x] = 0;

    scan_retry:
      scsi_cmd (inquiry, 6, image, 256);

      scsi_cmd (testunit, 6, NULL, 0);

      area[7] = ((unsigned int) (x_pos * 8)) & 0xff;
      area[8] = (((unsigned int) (x_pos * 8)) & 0xff00) >> 8;
      area[9] = ((unsigned int) (y_pos * 8)) & 0xff;
      area[10] = (((unsigned int) (y_pos * 8)) & 0xff00) >> 8;

      area[11] = ((unsigned int) ((x_pos + x_size) * 8)) & 0xff;
      area[12] = (((unsigned int) ((x_pos + x_size) * 8)) & 0xff00) >> 8;
      area[13] = ((unsigned int) ((y_pos + y_size) * 8)) & 0xff;
      area[14] = (((unsigned int) ((y_pos + y_size) * 8)) & 0xff00) >> 8;

      scsi_cmd (area, 15, NULL, 0);

      mode[7] = (resolution / 3) & 0xff;
      mode[8] = brightness & 0xff;
      mode[9] = contrast & 0xff;
      mode[10] = grain & 0xff;
      mode[11] = speed & 0xff;
      mode[14] = ((unsigned int) ((x_size + x_pos) * 8)) & 0xff;
      mode[15] = (((unsigned int) ((x_size + x_pos) * 8)) & 0xff00) >> 8;
      scsi_cmd (mode, 16, NULL, 0);

      scsi_cmd (lookup, 10, NULL, 0);

      mask = color ? (run == 0 ? 0x08 : (run == 1 ? 0x10 : 0x18)) : 0;
      if (color)
	rgb[run] = slice;

      start[4] = 0x41 + mask;
      scsi_cmd (start, 6, NULL, 0);

      line = 0;

      retry = 0;

      do
	{
	  scsi_cmd (status, 6, state, 6);
	  busy = state[0];
	  bpl = state[1] + state[2] * 256;
	  lib = state[3] + state[4] * 256 + state[5] * 65536;

	  TRACE (2, "rdy %d bpl(%d) %4d lib(%d) %8d\n", !busy, maxbpl, bpl, lib, maxline);

	  retry++;
	  if (busy)
	    sleep (retry);

	  if (retry > MAXRETRY)
	    break;
	}
      while (busy);

      if (bpl > 4095 || busy)
	{
	  if (bigretry < BIGRETRY)
	    {
	      bigretry++;
	      scsi_cmd (stop, 6, NULL, 0);

	      sleep (2);

	      TRACE (1, "bigretry %d\n", bigretry);

	      goto scan_retry;
	    }
	}

      backtrack[4] = (backtracking ? 0x82 : 0x80);
      scsi_cmd (backtrack, 6, NULL, 0);

      if (retry <= MAXRETRY)
	do
	  {
#ifdef SG_BIG_BUFF
	    ll = (mem_size - sizeof (struct sg_header)) / bpl;
	    if (ll > maxline - line)
	      ll = maxline - line;
	    if (ll < 1)
	      ll = 1;
#else
	    ll = 1;
#endif

	    readline[2] = ((ll & 0xff0000) >> 16) & 0xff;
	    readline[3] = ((ll & 0x00ff00) >> 8) & 0xff;
	    readline[4] = ((ll & 0x0000ff)) & 0xff;

	    scsi_cmd (readline, 6, image, bpl * ll);

	    for (y = 0; y < ll; y++)
	      for (x = 0; x < bpl && x < maxbpl; x++)
		slice[x + line * maxbpl + y * maxbpl] = image[x + y * bpl];

	    line += ll;
	  }
	while (line < maxline);

      TRACE (1, "lines processed %d, bpl processed %d, maxbpl %d, maxline %d\n", line, bpl, maxbpl, maxline);

      scsi_cmd (stop, 6, NULL, 0);

      for (i = 1; i <= MAXRETRY; i++)
	{
	  scsi_cmd (status, 6, state, 6);
	  if (!state[0])
	    break;
	  sleep (i);
	}

      scsi_cmd (stop, 6, NULL, 0);

      if (retry > MAXRETRY)
	ERROR ("scanner timed out\n");

      if (bpl > 4095)
	ERROR ("input line too long (%d, allowed 4095) - use lower resolution or x-size\n", bpl);
    }

  if (strcmp (outputfile, "-"))
    {
      if ((stream = fopen (outputfile, "w")) == NULL)
	ERROR ("cannot open <%s>: %s\n", outputfile, strerror (errno));
    }
  else
    stream = stdout;

  fprintf (stream, "P6\n#CREATOR: muscan Version %s\n%d %d\n255\n", VERSION, maxbpl, maxline);

  for (y = 0; y < maxline; y++)
    for (x = 0; x < maxbpl; x++)
      {
	unsigned char pixel[3];
	int ofs;

	ofs = x + y * maxbpl;

	pixel[0] = color ? rgb[0][ofs] : slice[ofs];
	pixel[1] = color ? rgb[1][ofs] : slice[ofs];
	pixel[2] = color ? rgb[2][ofs] : slice[ofs];

	fwrite (pixel, 3, 1, stream);
      }

  if (strcmp (outputfile, "-"))
    fclose (stream);

  free (slice);

  for (run = 0; run < 3; run++)
    if (rgb[run])
      free (rgb[run]);
}

/*******************************************
 * Main procedure                          *
 *******************************************/
void main
  (
    int argc,			/* I: argument count */
    char *argv[]		/* I: arguments */
)
{
  parse_commandline (argc, argv);

  open_scanner ();

  scan_image ();

  close_scanner ();
}

/*
   End of file
 */
