/*

   xmuscan - X11 frontend to the scanner driver for Mustek Paragon 6000CX scanner
   Copyright (C) 1995 Torsten Eichner

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

   email:       eichner@rhrk.uni-kl.de
   papermail:   Torsten Eichner, Kampshardtweg 1, D-57587 Birken-Honigsessen, Germany

 */

/*
   Start of file: xmuscan.c
 */

#include <X11/Intrinsic.h>
#include <X11/Shell.h>
#include <X11/StringDefs.h>

#include <X11/Core.h>
#include <X11/Xaw/AsciiText.h>
#include <X11/Xaw/Command.h>
#include <X11/Xaw/Label.h>
#include <X11/Xaw/Paned.h>
#include <X11/Xaw/Toggle.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <unistd.h>

#define X_IMAGE 9*9
#define Y_IMAGE 13*9

static Widget shell, drawarea;
static Widget mainwnd, commandbox;
static Widget wquit, wscan, wpreview, wabout;
static Widget lres, tres, bres;
static Widget lbri, tbri, bbri;
static Widget lcon, tcon, bcon;
static Widget total, wcolor;
static Widget wimage, wstate;
static Pixel color_array[64];
static unsigned char image[X_IMAGE * Y_IMAGE];
static Pixmap pixmap = 0;

static int res = 100;
static float xp = 0.0;
static float yp = 0.0;
static float xs = 9.0;
static float ys = 13.0;
static int color = 0;
static int contrast = 12;
static int brightness = 12;
static int imagestate = 1;

static void usage (char *);
static void parse_commandline (int, char *[]);
static void display_image (void);
static void preview (void);
static void scan (void);
static void cmd_cb (Widget, XtPointer, XtPointer);
static void value_cb (Widget, XtPointer, XtPointer);
static void popdown_cb (Widget, XtPointer, XtPointer);
static void update_values (void);
static void window_size (Widget, XEvent *);
static void laststate (char *);

/*******************************************
 * Set state window                        *
 *******************************************/
static void laststate
  (
    char *state			/* I: state info */
)
{
  XtVaSetValues (wstate, XtNlabel, state, NULL);
}

/*******************************************
 * Mouse in window                         *
 *******************************************/
static void window_size
  (
    Widget w,			/* I: widget */
    XEvent * event		/* I: calling event */
)
{
  int x, y;
  char buffer[1024];
  static int x1, x2, y1, y2;

  switch (event->type)
    {
    case ButtonPress:
      x = event->xbutton.x / 2;
      y = event->xbutton.y / 2;

      if (x > X_IMAGE)
	x = X_IMAGE;
      if (y > Y_IMAGE)
	y = Y_IMAGE;

      x1 = x;
      y1 = y;
      x2 = x;
      y2 = y;
      break;

    case ButtonRelease:
      break;

    case MotionNotify:
      x = event->xmotion.x / 2;
      y = event->xmotion.y / 2;

      sprintf (buffer, "Pointer at position %2.2f:%2.2f", x / 9.0, y / 9.0);

      laststate (buffer);

      if (!(event->xmotion.state & Button1Mask))
	return;

      if (x > X_IMAGE)
	x = X_IMAGE;
      if (y > Y_IMAGE)
	y = Y_IMAGE;

      x2 = (x < x1 ? x1 : x);
      y2 = (y < y1 ? y1 : y);
      break;
    }

  xp = x1 / 9.0;
  yp = y1 / 9.0;

  xs = (x2 - x1) / 9.0;
  ys = (y2 - y1) / 9.0;

  display_image ();
  update_values ();
}

/*******************************************
 * Print usage                             *
 *******************************************/
static void usage
  (
    char *text			/* I: erroneous argument */
)
{
  printf ("xmuscan V%s\n", VERSION);
  printf ("Syntax: xmuscan [-h] [-v] [-?] [X-Options]\n");
  printf ("Options:\n");
  printf (" -h : print this help\n");
  printf (" -v : show version\n");
  printf (" -? : print this help\n");

  if (text)
    printf ("Erroneous argument: <%s>\n", text);

  if (text)
    exit (-1);
  else
    exit (0);
}

/*******************************************
 * Parse commandline                       *
 *******************************************/
static void parse_commandline
  (
    int argc,			/* I: argument count */
    char *argv[]		/* I: arguments */
)
{
  int i;

  for (i = 1; i < argc; i++)
    if (argv[i][0] == '-')
      switch (argv[i][1])
	{
	case 'h':		/* help */
	  usage (NULL);
	  break;

	case 'v':		/* version */
	  printf ("xmuscan V%s - compiled by %s on %s\n", VERSION, USER, HOST);
	  printf ("Copyright (C) 1995 by Torsten Eichner, Kampshardtweg 1, D-57587 Birken-Honigsessen\n");
	  printf ("E-mail: eichner@rhrk.uni-kl.de\n");
	  printf ("This program is distributed under the GNU GPL - please refer to it for further information.\n");
	  break;

	case '?':		/* help */
	  usage (NULL);
	  break;

	default:
	  usage (argv[i]);
	}
    else
      usage (argv[i]);
}

/*******************************************
 * Display image                           *
 *******************************************/
static void display_image
  (
    void
)
{
  int x, y;
  int x1, x2, y1, y2;
  GC gc;
  XGCValues values;
  static Pixmap tmp = 0;

  values.foreground = values.background = color_array[0];
  gc = XCreateGC (XtDisplay (shell), pixmap, GCForeground | GCBackground, &values);

  if (imagestate)
    {
      if (tmp)
	XFreePixmap (XtDisplay (shell), tmp);

      tmp = XCreatePixmap (XtDisplay (shell), RootWindowOfScreen (XtScreen (shell)), X_IMAGE * 2, Y_IMAGE * 2, DefaultDepthOfScreen (XtScreen (shell)));

      for (y = 0; y < Y_IMAGE; y++)
	for (x = 0; x < X_IMAGE; x++)
	  {
	    int iv;

	    iv = image[x + y * X_IMAGE];

	    iv = (iv / 4 > 63 ? 63 : iv / 4);
	    values.foreground = color_array[iv];

	    XChangeGC (XtDisplay (shell), gc, GCForeground, &values);

	    XFillRectangle (XtDisplay (shell), tmp, gc, 2 * x, 2 * y, 2, 2);
	  }
      laststate ("Created fresh image");
    }

  XCopyArea (XtDisplay (shell), tmp, pixmap, DefaultGCOfScreen (XtScreen (shell)), 0, 0, X_IMAGE * 2, Y_IMAGE * 2, 0, 0);

  x1 = ((int) (xp * 9)) * 2;
  x2 = ((int) (xs * 9)) * 2 + 2;
  y1 = ((int) (yp * 9)) * 2;
  y2 = ((int) (ys * 9)) * 2 + 2;

  values.foreground = color_array[0];
  XChangeGC (XtDisplay (shell), gc, GCForeground, &values);
  XDrawRectangle (XtDisplay (shell), pixmap, gc, x1, y1, x2, y2);

  values.foreground = color_array[32];
  XChangeGC (XtDisplay (shell), gc, GCForeground, &values);
  XDrawRectangle (XtDisplay (shell), pixmap, gc, x1 + 1, y1 + 1, x2 - 2, y2 - 2);

  XFreeGC (XtDisplay (shell), gc);

  XCopyArea (XtDisplay (shell), pixmap, XtWindow (drawarea), DefaultGCOfScreen (XtScreen (shell)), 0, 0, X_IMAGE * 2, Y_IMAGE * 2, 0, 0);
  XtVaSetValues (drawarea, XtNwidth, X_IMAGE * 2, XtNheight, Y_IMAGE * 2, NULL);

  imagestate = 0;
}

/*******************************************
 * Preview image                           *
 *******************************************/
static void preview
  (
    void
)
{
  int pd[2];
  int pid;
  int status;
  char buffer[43 + X_IMAGE * Y_IMAGE * 3];
  int n, x, y, c;

  if (pipe (pd) != 0)
    XtError ("cannot create pipe");

  laststate ("Trying to scan image...");

  if ((pid = fork ()) == 0)
    {
      close (1);
      close (pd[0]);

      dup (pd[1]);

      execlp ("muscan", "muscan", "-b", "-r", "9", "-x", "9", "-y", "13", NULL);

      exit (-1);
    }

  close (pd[1]);

  c = 43 + X_IMAGE * Y_IMAGE * 3;

  while (c > 0)
    {
      n = read (pd[0], buffer + 43 + X_IMAGE * Y_IMAGE * 3 - c, c);

      c -= n;

      if (n == 0)
	break;
    }

  for (x = 0; x < X_IMAGE; x++)
    for (y = 0; y < Y_IMAGE; y++)
      image[x + y * X_IMAGE] = ((c == 0) ? buffer[43 + x * 3 + y * 3 * X_IMAGE] : 0);

  imagestate = 1;

  display_image ();

  waitpid (pid, &status, 0);
  close (pd[0]);

  if (status != 0)
    laststate ("Error in reading image!");
}

/*******************************************
 * Scan image                              *
 *******************************************/
static void scan
  (
    void
)
{
  char buffer[4096];

  if (color)
    sprintf (buffer, "muscan -r %d -x %f -y %f -l %f -u %f %s -p %d:%d:0:1 | xv - &", res, xs, ys, xp, yp, color ? "-c" : "", brightness, contrast);
  else
    sprintf (buffer, "muscan -b -r %d -x %f -y %f -l %f -u %f %s -p %d:%d:0:1 | xv - &", res, xs, ys, xp, yp, color ? "-c" : "", brightness, contrast);

  system (buffer);

  laststate ("Scanning image...");
}

/*******************************************
 * Command callback                        *
 *******************************************/
static void cmd_cb
  (
    Widget widget,		/* I: widget */
    XtPointer cld,		/* I: client data */
    XtPointer cad		/* I: call data */
)
{
  static Widget about = 0;
  Dimension w, h;

  switch ((int) cld)
    {
    case 0:			/* scan */
      value_cb (0, NULL, NULL);
      scan ();
      break;

    case 1:			/* preview */
      value_cb (0, NULL, NULL);
      preview ();
      break;

    case 2:			/* about */
      value_cb (0, NULL, NULL);

      if (!about)
	{
	  Widget dlg, cmd, lbl;
	  char abouttext[1024];
	  Position x, y;
	  Dimension w, h;

	  sprintf (abouttext, "xmuscan - X11 frontend for muscan V%s\nCopyright (C) 1995 by Torsten Eichner\neichner@rhrk.uni-kl.de\nCompiled by %s on %s\n", VERSION, USER, HOST);

	  about = XtVaCreatePopupShell (NULL, transientShellWidgetClass, shell, NULL);
	  dlg = XtVaCreateManagedWidget (NULL, panedWidgetClass, about, NULL);
	  lbl = XtVaCreateManagedWidget (NULL, labelWidgetClass, dlg, XtNshowGrip, False, NULL);
	  cmd = XtVaCreateManagedWidget ("ok", commandWidgetClass, dlg, XtNskipAdjust, True, NULL);

	  XtVaGetValues (shell, XtNwidth, &w, XtNheight, &h, NULL);
	  XtTranslateCoords (shell, (Position) w / 2, (Position) h / 2, &x, &y);

	  XtVaSetValues (about, XtNx, x, XtNy, y, NULL);

	  XtVaSetValues (lbl, XtNlabel, abouttext, NULL);

	  XtAddCallback (cmd, XtNcallback, popdown_cb, (XtPointer) about);
	}

      XtManageChild (about);

      XtVaGetValues (about, XtNwidth, &w, XtNheight, &h, NULL);
      XtVaSetValues (about, XtNmaxWidth, w, XtNminWidth, w, XtNmaxHeight, h, XtNminHeight, h, NULL);

      XtPopup (about, XtGrabNone);

      laststate ("Displaying program info");
      break;

    case 3:			/* quit */
      exit (0);
      break;
    }
}

/*******************************************
 * Popdown callback                        *
 *******************************************/
static void popdown_cb
  (
    Widget w,			/* I: widget */
    XtPointer cld,		/* I: client data */
    XtPointer cad		/* I: call data */
)
{
  XtPopdown ((Widget) cld);
}

/*******************************************
 * Value callback                          *
 *******************************************/
static void value_cb
  (
    Widget w,			/* I: widget */
    XtPointer cld,		/* I: client data */
    XtPointer cad		/* I: call data */
)
{
  char *buffer;
  int val;

  XtVaGetValues (tres, XtNstring, &buffer, NULL);
  if (sscanf (buffer, "%d", &val) != 1 || val < 3 || val > 300)
    val = 100;
  res = val;
  XtVaGetValues (tbri, XtNstring, &buffer, NULL);
  if (sscanf (buffer, "%d", &val) != 1 || val < 0 || val > 24)
    val = 12;
  brightness = val;
  XtVaGetValues (tcon, XtNstring, &buffer, NULL);
  if (sscanf (buffer, "%d", &val) != 1 || val < 0 || val > 24)
    val = 12;
  contrast = val;

  XtVaGetValues (wcolor, XtNstate, &val, NULL);
  color = val;

  update_values ();
}

/*******************************************
 * Update values                           *
 *******************************************/
static void update_values
  (
    void
)
{
  char buffer[1024];

  sprintf (buffer, "%d", res);
  XtVaSetValues (tres, XtNstring, buffer, NULL);
  sprintf (buffer, "%d", contrast);
  XtVaSetValues (tcon, XtNstring, buffer, NULL);
  sprintf (buffer, "%d", brightness);
  XtVaSetValues (tbri, XtNstring, buffer, NULL);
  sprintf (buffer, "Image is %2.2fx%2.2f at position %2.2f:%2.2f", xs, ys, xp, yp);
  XtVaSetValues (wimage, XtNlabel, buffer, NULL);
  sprintf (buffer, "Total bytes per scan: %10d", (int) (xs * ys * res * res * (color ? 3 : 1)));
  XtVaSetValues (total, XtNlabel, buffer, NULL);
}

/*******************************************
 * Main procedure                          *
 *******************************************/
void main
  (
    int argc,			/* I: argument count */
    char *argv[]		/* I: arguments */
)
{
  XtAppContext context;
  XColor color;
  static Colormap colormap;
  GC gc;
  XGCValues values;
  int i;
  static XtActionsRec action[] =
  {
    {"DisplayImage", (XtActionProc) display_image},
    {"WindowSize", (XtActionProc) window_size},
  };

  /* initialize application */
  parse_commandline (argc, argv);

  /* create frontend */
  shell = XtAppInitialize (&context, "Xmuscan", NULL, 0, &argc, argv, NULL, NULL, 0);

  mainwnd = XtVaCreateManagedWidget (NULL, panedWidgetClass, shell, XtNorientation, XtorientHorizontal, NULL);
  drawarea = XtVaCreateManagedWidget (NULL, coreWidgetClass, mainwnd, XtNshowGrip, False, XtNheight, Y_IMAGE * 2, XtNwidth, X_IMAGE * 2, NULL);
  commandbox = XtVaCreateManagedWidget (NULL, panedWidgetClass, mainwnd, NULL);

  wscan = XtVaCreateManagedWidget ("scan", commandWidgetClass, commandbox, XtNshowGrip, False, NULL);
  wpreview = XtVaCreateManagedWidget ("preview", commandWidgetClass, commandbox, XtNshowGrip, False, NULL);
  wabout = XtVaCreateManagedWidget ("about", commandWidgetClass, commandbox, XtNshowGrip, False, NULL);
  wquit = XtVaCreateManagedWidget ("quit", commandWidgetClass, commandbox, XtNshowGrip, False, NULL);

  bres = XtVaCreateManagedWidget (NULL, panedWidgetClass, commandbox, XtNshowGrip, False, XtNorientation, XtorientHorizontal, NULL);
  lres = XtVaCreateManagedWidget ("resolution", labelWidgetClass, bres, XtNshowGrip, False, NULL);
  tres = XtVaCreateManagedWidget (NULL, asciiTextWidgetClass, bres, XtNeditType, XawtextEdit, NULL);

  bcon = XtVaCreateManagedWidget (NULL, panedWidgetClass, commandbox, XtNshowGrip, False, XtNorientation, XtorientHorizontal, NULL);
  lcon = XtVaCreateManagedWidget ("contrast", labelWidgetClass, bcon, XtNshowGrip, False, NULL);
  tcon = XtVaCreateManagedWidget (NULL, asciiTextWidgetClass, bcon, XtNeditType, XawtextEdit, NULL);

  bbri = XtVaCreateManagedWidget (NULL, panedWidgetClass, commandbox, XtNshowGrip, False, XtNorientation, XtorientHorizontal, NULL);
  lbri = XtVaCreateManagedWidget ("brightness", labelWidgetClass, bbri, XtNshowGrip, False, NULL);
  tbri = XtVaCreateManagedWidget (NULL, asciiTextWidgetClass, bbri, XtNeditType, XawtextEdit, NULL);

  wcolor = XtVaCreateManagedWidget ("color", toggleWidgetClass, commandbox, NULL);

  wimage = XtVaCreateManagedWidget (NULL, labelWidgetClass, commandbox, XtNshowGrip, False, NULL);

  total = XtVaCreateManagedWidget (NULL, labelWidgetClass, commandbox, XtNshowGrip, False, NULL);
  wstate = XtVaCreateManagedWidget (NULL, labelWidgetClass, commandbox, XtNshowGrip, False, NULL);

  bzero (image, sizeof (image));
  XtRealizeWidget (shell);

  XtAddCallback (wscan, XtNcallback, cmd_cb, (XtPointer) 0);
  XtAddCallback (wpreview, XtNcallback, cmd_cb, (XtPointer) 1);
  XtAddCallback (wabout, XtNcallback, cmd_cb, (XtPointer) 2);
  XtAddCallback (wquit, XtNcallback, cmd_cb, (XtPointer) 3);

  XtAddCallback (wcolor, XtNcallback, value_cb, NULL);

  /* allocate 64 colors */
  colormap = DefaultColormapOfScreen (XtScreen (shell));

  for (i = 0; i < 64; i++)
    {
      color.red = color.blue = color.green = (i << 10);

      if (!XAllocColor (XtDisplay (shell), colormap, &color))
	break;

      color_array[i] = color.pixel;
    }

  if (i < 64)
    XtError ("cannot allocate 64 colors");

  /* create pixmap */
  pixmap = XCreatePixmap (XtDisplay (shell), RootWindowOfScreen (XtScreen (shell)), X_IMAGE * 2, Y_IMAGE * 2, DefaultDepthOfScreen (XtScreen (shell)));

  values.background = values.foreground = color_array[0];
  gc = XCreateGC (XtDisplay (shell), pixmap, GCForeground | GCBackground, &values);

  XFillRectangle (XtDisplay (shell), pixmap, gc, 0, 0, X_IMAGE * 2, Y_IMAGE * 2);

  XFreeGC (XtDisplay (shell), gc);

  display_image ();

  update_values ();

  XtAppAddActions (context, action, XtNumber (action));

  XtOverrideTranslations (drawarea, XtParseTranslationTable ("<Expose>: DisplayImage()\n<Btn1Down>: WindowSize()\n<Btn1Motion>: WindowSize()\n<Btn1Up>: WindowSize()\n<Motion>: WindowSize()\n"));

  /* application loop */
  XtAppMainLoop (context);
}

/*
   End of file
 */
